/*
   Copyright (c) 1995-1996 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute this software for
   any purpose and without fee is hereby granted, provided that this
   copyright and permission notice appear on all copies of the
   software and supporting documentation, the name of Cisco Systems,
   Inc. not be used in advertising or publicity pertaining to
   distribution of the program without specific prior permission, and
   notice be given in supporting documentation that modification,
   copying and distribution is by permission of Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/

#include <strings.h>
#include <stdlib.h>
#include "tac_plus.h"
#include "expire.h"
#include "md5.h"

#ifdef ARAP_DES
#include "arap_des.h"
#endif

/* internal state variables */
#define STATE_AUTHEN_START   0	/* no requests issued */
#define STATE_AUTHEN_GETUSER 1	/* username has been requested */
#define STATE_AUTHEN_GETPASS 2	/* password has been requested */

struct private_data {
    char password[MAX_PASSWD_LEN + 1];
    int state;
};

static void chap_verify();
static void arap_verify();
static void pap_verify();
static void login();

/*
 * Default tacacs login authentication function. Wants a username
 * and a password, and tries to verify them.
 *
 * Choose_authen will ensure that we already have a username before this
 * gets called.
 *
 * We will query for a password and keep it in the method_data.
 *
 * Any strings returned via pointers in authen_data must come from the
 * heap. They will get freed by the caller.
 *
 * Return 0 if data->status is valid, otherwise 1
 */

int
default_fn(data)
struct authen_data *data;
{
    struct private_data *p;
    char *name = data->NAS_id->username;

    p = (struct private_data *) data->method_data;

    /* An abort has been received. Clean up and return */
    if (data->flags & TAC_PLUS_CONTINUE_FLAG_ABORT) {
	if (data->method_data)
	    free(data->method_data);
	data->method_data = NULL;
	return (1);
    }

    /* Initialise method_data if first time through */
    if (!p) {
	p = (struct private_data *) tac_malloc(sizeof(struct private_data));
	bzero(p, sizeof(struct private_data));
	data->method_data = p;
	p->state = STATE_AUTHEN_START;
    }

    if (STREQ(name, DEFAULT_USERNAME)) {
	/* Never authenticate this user. It's for authorization only */
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
/*	if (debug) { */
	    report(LOG_INFO,
		   "authentication query for '%s' %s from %s rejected",
		   name && name[0] ? name : "unknown",
		   session.port, session.peer);
/*	} */
	return(0);
    }

    if (data->action != TAC_PLUS_AUTHEN_LOGIN) {
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
    } else {
	switch(data->type) {
	case TAC_PLUS_AUTHEN_TYPE_CHAP:
	    /* set status inside chap_verify */
	    chap_verify(data);

	    if (debug ||!(data->status == TAC_PLUS_AUTHEN_STATUS_PASS)) {
		report(LOG_INFO, "chap-login query for '%s' %s from %s %s",
		       name && name[0] ? name : "unknown",
		       session.port,  session.peer,
		       (data->status == TAC_PLUS_AUTHEN_STATUS_PASS) ?
		       "accepted" : "rejected");
	    }
	    break;

	case TAC_PLUS_AUTHEN_TYPE_ARAP:
	    /* set status inside arap_verify */
	    arap_verify(data);

	    if (debug || !(data->status == TAC_PLUS_AUTHEN_STATUS_PASS)) {
		report(LOG_INFO, "arap query for '%s' %s from %s %s",
		       name && name[0] ? name : "unknown",
		       session.port, session.peer,
		       (data->status == TAC_PLUS_AUTHEN_STATUS_PASS) ?
		       "accepted" : "rejected");
	    }
	    break;

	case TAC_PLUS_AUTHEN_TYPE_PAP:
	    pap_verify(data);

	    if (debug || !(data->status == TAC_PLUS_AUTHEN_STATUS_PASS)) {
		report(LOG_INFO, "pap-login query for '%s' %s from %s %s",
		       name && name[0] ? name : "unknown",
		       session.port,
		       session.peer,
		       (data->status == TAC_PLUS_AUTHEN_STATUS_PASS) ?
		       "accepted" : "rejected");
	    }
	    break;

	case TAC_PLUS_AUTHEN_TYPE_ASCII:
	    login(data, p);
	    switch(data->status) {
	    case TAC_PLUS_AUTHEN_STATUS_GETPASS:
	    case TAC_PLUS_AUTHEN_STATUS_GETUSER:
	    case TAC_PLUS_AUTHEN_STATUS_GETDATA:
		/* Authentication still in progress. More data required */
		return(0);

	    default:
		/* Authentication finished */
		if (debug || !(data->status == TAC_PLUS_AUTHEN_STATUS_PASS))
		    report(LOG_INFO, "login query for '%s' %s from %s %s",
			   name && name[0] ? name : "unknown",
			   session.port,
			   session.peer,
			   (data->status == TAC_PLUS_AUTHEN_STATUS_PASS) ?
			   "accepted" : "rejected");
	        if(data->status == TAC_PLUS_AUTHEN_STATUS_PASS && !data->server_msg)
	         {
	          char *msg;
	          if((msg = cfg_get_pvalue (name, TAC_IS_USER, S_message, TAC_PLUS_RECURSE)) != NULL)
	           {
	            data->server_msg =
#ifdef MORE_SUBST
                                       substitute2(msg, data);
#else
	                               tac_strdup(msg);
#endif
                   }
	         }
	    }
	    break;

	default:
	    data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	    break;
	}
    }

    if (data->method_data)
	free(data->method_data);
    data->method_data = NULL;

    switch (data->status) {
    case TAC_PLUS_AUTHEN_STATUS_ERROR:
    case TAC_PLUS_AUTHEN_STATUS_FAIL:
    case TAC_PLUS_AUTHEN_STATUS_PASS:
	return (0);

    default:
	report(LOG_ERR, "%s %s: default_fn set bogus status value %d",
	       session.peer, session.port, data->status);
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return (0);
    }
}

/* Do a login requiring a username & password. We already know the
 * username. We may return GETPASS to get a password if we need it.
 * The password will be stored in the private data
 *
 */

static void
login(data, p)
    struct authen_data *data;
    struct private_data *p;
{
    char *name, *passwd;
    int pwlen;

#ifdef NONE_LOGIN
    char *secret;
    int nopassword = 0;
#endif

    name = data->NAS_id->username;

    if (!name[0]) {
	/* something awful has happened. Give up and die */
	report(LOG_ERR, "%s %s: no username for login",
	       session.peer, session.port);
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return;
    }

    /* Do we have a password? */
    passwd = p->password;

#ifdef NONE_LOGIN
    if ((secret = cfg_get_login_secret(name, TAC_PLUS_RECURSE)) != NULL)
     {
      if(STREQ(secret, "permit"))
        nopassword = 1;
      else if(STREQ(secret, "deny"))
        nopassword = 2;
      else if(STREQ(secret, "none"))
        nopassword = -1;
     }

    if (nopassword)
      ;
    else
#endif

    if (!passwd[0]) {

	/* no password yet. Either we need to ask for one and expect
	 * to get called again when it's supplied, or we already asked
	 * for one and we should have a reply.
	 */

	switch (p->state) {
	case STATE_AUTHEN_GETPASS:
	    /* We already asked for a password. This should be
	       the reply */
	    pwlen = MIN((int)strlen(data->client_msg), MAX_PASSWD_LEN);
	    strncpy(passwd, data->client_msg, pwlen);
	    passwd[pwlen] = '\0';
	    break;

	default:
	    data->flags = TAC_PLUS_AUTHEN_FLAG_NOECHO;
	    fr_fi(data->server_msg, tac_strdup("Password: "));
	    data->status = TAC_PLUS_AUTHEN_STATUS_GETPASS;
	    p->state = STATE_AUTHEN_GETPASS;
	    return;
	}
    }

    /* Now we have a username and password. Try validating */

    /* Assume the worst */
    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

#ifdef NONE_LOGIN
    switch(nopassword)
     {
      case 0:			/* password */
      case -1:			/* none     */
        break;			/* standard procedure */
      case 1:			/* permit - set status and no procedure  */
        data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
      default:			/* deny - no procedure */
        return;
     }
#endif

    verify(name, passwd, data, TAC_PLUS_RECURSE);
    return;
}

/*
 * Process an inbound PAP login. The username & password should be in
 * the START packet.
 */

static void
pap_verify(data)
    struct authen_data *data;
{
    char *name, *passwd;

    name = data->NAS_id->username;

    if (!name[0]) {
	/* something awful has happened. Give up and die */
	report(LOG_ERR, "%s %s: no username for inbound PAP login",
	       session.peer, session.port);
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return;
    }

    /* get the password */
    passwd = tac_malloc(data->client_dlen + 1);
    bcopy(data->client_data, passwd, data->client_dlen);
    passwd[data->client_dlen] = '\0';

    /* Assume the worst */
    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
    verify(name, passwd, data, TAC_PLUS_RECURSE);
    free(passwd);
}


/* Verify the challenge and id against the response by looking up the
 * chap secret in the config file. Set data->status appropriately.
 */
static void
chap_verify(data)
struct authen_data *data;
{
    char *name, *secret, *chal, digest[MD5_LEN];
    char *exp_date, *p;
    u_char *mdp;
    char id;
    int chal_len, inlen;
    MD5_CTX mdcontext;

    if (!(char) data->NAS_id->username[0]) {
	report(LOG_ERR, "%s %s: no username for chap_verify",
	       session.peer, session.port);
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return;
    }

    name = data->NAS_id->username;

    id = data->client_data[0];

    chal_len = data->client_dlen - 1 - MD5_LEN;
    if (chal_len < 0) {
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return;
    }

    if (debug & DEBUG_AUTHEN_FLAG) {
	report(LOG_DEBUG, "%s %s: chap user=%s, id=%d chal_len=%d",
	       session.peer, session.port, name, (int)id, chal_len);

	/* report_hex(LOG_DEBUG, (u_char *)data->client_data + 1, chal_len); */
    }

    /* Assume failure */
    data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;

    /* Get the secret */
    secret = cfg_get_chap_secret(name, TAC_PLUS_RECURSE);

    /* If there is no chap password for this user, see if there is
       a global password for her that we can use */
    if (!secret) {
	secret = cfg_get_global_secret(name, TAC_PLUS_RECURSE);
    }

    if (!secret) {
	/* No secret. Fail */
	if (debug & DEBUG_AUTHEN_FLAG) {
	    report(LOG_DEBUG, "%s %s: No chap or global secret for %s",
		   session.peer, session.port, name);
	}
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return;
    }

    p = tac_find_substring("cleartext ", secret);
    if (!p) {
	report(LOG_ERR, "%s %s: %s chap secret %s is not cleartext",
	       session.peer, session.port, name, secret);
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return;
    }
    secret = p;

    /*
     * We now have the secret, the id, and the challenge value.
     * Put them all together, and run them through the MD5 digest
     * algorithm. */

    inlen = sizeof(u_char) + strlen(secret) + chal_len;
    mdp = (u_char *)tac_malloc(inlen);
    mdp[0] = id;
    bcopy(secret, &mdp[1], strlen(secret));
    chal = data->client_data + 1;
    bcopy(chal, mdp + strlen(secret) + 1, chal_len);
    MD5Init(&mdcontext);
    MD5Update(&mdcontext, mdp, inlen);
    MD5Final((u_char *)digest, &mdcontext);
    free(mdp);

    /*
     * Now compare the received response value with the just
     * calculated digest value.  If they are equal, it's a pass,
     * otherwise it's a failure */

    if (bcmp(digest, data->client_data+1+chal_len, MD5_LEN)) {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
    } else {
	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
    }

    exp_date = cfg_get_expires(name, TAC_PLUS_RECURSE);
    set_expiration_status(exp_date, data);
#if defined(DENY_FOR_PPP) || defined(CRON_FOR_PPP) || defined(DBF_FOR_PPP)
    sys_adm_verify(name, 0, data, TAC_PLUS_RECURSE);
#endif
}


/*
 * Force the "parity" bit to zero on a password before passing it to
 * des. This is not documented anywhere. (I believe forcing the parity
 * to zero reduces the integrity of the encrypted keys but this is
 * what Apple chose to do).
 */
void pw_bitshift(pw)
char *pw;
{
    int i;
    unsigned char pws[8];

    /* key is 0 padded */
    for (i = 0; i < 8; i++)
        pws[i] = 0;

    /* parity bit is always zero (this seem bogus) */
    for (i = 0; i < 8 && pw[i]; i++)
        pws[i] = pw[i] << 1;

    bcopy(pws, pw, 8);
}


static void
arap_verify(data)
struct authen_data *data;
{
    char nas_chal[8], r_chal[8], r_resp[8], secret[8];
    char *name, *cfg_secret, *exp_date, *p;

    if (!(char) data->NAS_id->username[0]) {
	report(LOG_ERR, "%s %s: no username for arap_verify",
	       session.peer, session.port);
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return;
    }

    name = data->NAS_id->username;

    bcopy(data->client_data, nas_chal, 8);
    bcopy(data->client_data+8, r_chal, 8);
    bcopy(data->client_data+8+8, r_resp, 8);

    /* Assume failure */
    data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;

    /* Get the secret */
    cfg_secret = cfg_get_arap_secret(name, TAC_PLUS_RECURSE);

    /* If there is no arap password for this user, see if there is
       a global password for her that we can use */
    if (!cfg_secret) {
	cfg_secret = cfg_get_global_secret(name, TAC_PLUS_RECURSE);
    }

    if (!cfg_secret) {
	/* No secret. Fail */
	if (debug & DEBUG_AUTHEN_FLAG) {
	    report(LOG_DEBUG, "%s %s: No arap or global secret for %s",
		   session.peer, session.port, name);
	}
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return;
    }

    p = tac_find_substring("cleartext ", cfg_secret);
    if (!p) {
	report(LOG_ERR, "%s %s: %s arap secret %s is not cleartext",
	       session.peer, session.port, name, cfg_secret);
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return;
    }

    /* need to allocate 8 bytes for secret, even if it's actually shorter */
    bzero(secret, sizeof(secret));
    strcpy(secret, p);

    pw_bitshift(secret);

#ifdef ARAP_DES
    des_init(0);
    des_setkey(secret);
    des_endes(nas_chal);
    des_done();
#endif /* ARAP_DES */

    /*
     * Now compare the remote's response value with the just
     * calculated one value.  If they are equal, it's a pass,
     * otherwise it's a failure
     */

    if (bcmp(nas_chal, r_resp, 8)) {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
    } else {
	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
    }

#ifdef ARAP_DES
    /* Now calculate the response to the remote's challenge */
    des_init(0);
    des_setkey(secret);
    des_endes(r_chal);
    des_done();
#endif /* ARAP_DES */

    data->server_data = tac_malloc(8);
    data->server_dlen = 8;
    bcopy(r_chal, data->server_data, 8);

    exp_date = cfg_get_expires(name, TAC_PLUS_RECURSE);
    set_expiration_status(exp_date, data);
}
